// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// compress/decompress inverted list
// $Id: Coder.cc,v 0.9 1997-12-27 07:07:48+09 harada Exp $
//
extern "C" {
#include <unistd.h>
}
#include "Coder.h"

char *
Coder::pack(const unsigned int *src, int srcNum,
           char *dstbuf, size_t dstsize, size_t &resultsize){
    reset();
    bufp = dstbuf;
    for(int i = 0; i < srcNum; i++){
        if((size_t)(bufp - dstbuf) >= dstsize)
            return NULL; // compression fail
        delta_code(src[i]);
    }
    if(flush_need){ bit_flush(); }
    resultsize = bufp - dstbuf;

    return dstbuf;
}

char *
Coder::diffpack(const unsigned int *src, int srcNum,
                char *dstbuf, size_t dstsize, size_t &resultsize){
    reset();
    bufp = dstbuf;
    unsigned int prev = 0;
    for(int i = 0; i < srcNum; i++){
        if((size_t)(bufp - dstbuf) >= dstsize){
            resultsize = bufp - dstbuf;
            return NULL; // compression fail
        }
        delta_code(src[i] - prev);
        prev = src[i];
    }
    if(flush_need){ bit_flush(); }
    resultsize = bufp - dstbuf;

    return dstbuf;
}

extern "C"{
static int
compare_uint(const unsigned int *a, const unsigned int *b){
    return *a - *b;
}
}

void
Coder::sort(unsigned int *array, int arrayNum){
    qsort(array, arrayNum, sizeof(array[0]),
          (int (*)(const void *, const void *))compare_uint);
}

// src Τunpack˻Ȥ줿Ĺ֤
size_t
Coder::unpack(const char *src, unsigned int *dstbuf, int dstNum){
    reset();
    inbufp = src;
    for(int i = 0; i < dstNum; i++){
        dstbuf[i] = get_delta();
    }
    return inbufp - src;
}

size_t
Coder::unpack(const char *src, int srcLen, unsigned int *dstbuf){
    reset();
    inbufp = src;
    for(int i = 0; inbufp < src + srcLen; i++){
        dstbuf[i] = get_delta();
    }
    return inbufp - src;
}

size_t
Coder::diffunpack(const char *src, unsigned int *dstbuf, int dstNum){
    reset();
    inbufp = src;
    int prev = 0;
    for(int i = 0; i < dstNum; i++){
        prev += get_delta();
        dstbuf[i] = prev;
    }
    return inbufp - src;
}

int
Coder::diffunpack(const char *src, int srcLen,
                  unsigned int *dstbuf, int dstLen){
    reset();
    inbufp = src;
    int prev = 0;
    int i;
    for(i = 0; inbufp < src + srcLen && i < dstLen; i++){
        prev += get_delta();
        dstbuf[i] = prev;
    }
    return i;
}


// ------------------------------------------------------------
// private member functions

// bit I/O
int
Coder::bit_out(int bit){
    bit_out_byte = bit + (bit_out_byte << 1);
    if(bit_out_cnt == 7){
        *bufp++ = bit_out_byte;
        bit_out_cnt = 0; bit_out_byte = 0;
    }else
        bit_out_cnt++;
    return bit_out_cnt;
}

int
Coder::bit_in(){
    int bit;

    if(bit_in_filt == 1){
        bit_in_byte = *inbufp++;
        bit_in_filt = 0x80;
    }else
        bit_in_filt = bit_in_filt >> 1;
    bit = ((bit_in_byte & bit_in_filt) == 0) ? 0 : 1;
    return bit;
}

// filt ̤bit code 沽
void
Coder::fit_code(int code, int filt){
    int bit;
    filt = filt >> 1;
    while(filt > 0){
        bit = ((code & filt) == 0) ? 0 : 1;
        flush_need = bit_out(bit);
        filt = filt >> 1;
    }
}

// 沽
void
Coder::gamma_code(int code){
    int i, ret, leng;
    leng = i_log(code, ret);
    for(i = 1; i < leng; i++)
        flush_need = bit_out(0); // ޤ 0 
    fit_code(code, ret);         // 2ʿǽ
//  return code;
}

// 沽
void
Coder::delta_code(int code){
    int ret, leng;
    leng = i_log(code, ret);
    gamma_code(leng);           // Ĺ
    ret = ret >> 1;
    fit_code(code, ret);        // 2ʿǽ
}

// 沽
unsigned int
Coder::get_gamma(){
    int i, bit, leng = 0;
    unsigned int code = 1;

    while(bit_in() == 0) leng++; // Ϣ³0Ĺ
    for(i = 0; i < leng; i++){   // 2ʿɤ߹
        bit = bit_in();
        code = (code << 1) + bit;
    }
    return code;
}

// 沽
unsigned int
Coder::get_delta(){
    int i, bit, leng = 0;
    unsigned int code = 1;

    leng = get_gamma();         //  leng > 0
    leng--;
    for(i = 0; i < leng; i++){   // 2ʿɤ߹
        bit = bit_in();
        code = (code << 1) + bit;
    }
    return code;
}


#ifdef CODER_DEBUG
// g++ -O -g -DCODER_DEBUG Coder.cc
#include <iostream.h>

int main(){
#include "sample.cc"
    Coder coder;
    int found = sizeof(result)/sizeof(result[0]);
    size_t bufsize = sizeof(result);
    char *packedseq = new char[bufsize];
    size_t packedsize;
    
    coder.sort(result, found);
    if(NULL == coder.diffpack(result, found, packedseq,
                              bufsize, packedsize)){
        cerr << "Error!: insufficient buffer size" << endl;
        return 1;
    }
    cout << "before: " << sizeof(result) << "bytes ( ";
    cout << (sizeof(result) * 22) / 32 << ")." << endl;
    cout << " after: " << packedsize << "bytes\n";

    size_t restore[found];
    //    size_t unpacked = coder.diffunpack(packedseq, restore, found);
    size_t unpacked = coder.diffunpack(packedseq, packedsize, restore);
    cout << "unpack(): " << unpacked << " bytes used for unpacking\n";

    for(int j = 0; j < found; j++){
        cout << "result[] = " << result[j];
        cout << ", restore[] = " << restore[j] << endl;
    }
    
    delete []packedseq;
    
    return 0;
}
#endif
