// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Description.cc,v 0.9 1997-12-27 07:09:23+09 harada Exp $
//
extern "C" {
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <sys/param.h>          // ntohl(3)
#include <netinet/in.h>         // ntohl(3) (Linux, etc)
}
#include "config.h"
#include "Description.h"
#include "htmlutil.h"

void
HtmlDescription::getblock(int n, time_t &tm, size_t &sz, char *url,
                          char *author, char *title, char *digest){
    url[0] = author[0] = title[0] = digest[0] = '\0';
    tm = sz = 0;

    char buf[blksize + 1];
    file.seekg(blksize * n);
    file.read(buf, blksize);
    if(file.gcount() != (size_t)blksize){
        PRINTERRMSG("read error.");
        return;
    }
    buf[blksize] = '\0';

    sz = ntohl(*(size_t *)buf);
    tm = ntohl(*(time_t *)(buf + sizeof(size_t) + sizeof(time_t)));

    char *cp = buf + sizeof(size_t) + 2 * sizeof(time_t);
    strcpy(url, cp);

    while(*cp && cp < buf + blksize - 1) cp++;
    strcpy(author, ++cp);

    while(*cp && cp < buf + blksize - 1) cp++;
    strcpy(title, ++cp);

    while(*cp && cp < buf + blksize - 1) cp++;
    strcpy(digest, ++cp);
}

void
HtmlDescription::print(ostream &out, const Result &result){
    char url[blksize + 1];
    char author[blksize + 1];
    char title[blksize + 1];
    char digest[blksize + 1];
    time_t tm;
    size_t sz;

    getblock(result.docid, tm, sz, url, author, title, digest);
    out << "* docid = " << result.docid << endl;
    out << '"' << title << '"' << endl;
    out << "  URL: " << url;
    out << " (score: " << (int)result.weight;
    out << ", size: " << sz << " bytes)" << endl;
    out << "  Date: " << ctime(&tm);
    if(author[0]){
        out << "  Author: " << author << endl;
    }
    out << ">> " << digest << ".." << endl;
    out << endl; 
}

void
HtmlDescription::printHtml(ostream &out, const Result &result){
    char url[blksize + 1];
    char author[blksize + 1];
    char title[blksize + 1];
    char digest[blksize + 1];
    time_t tm;
    size_t sz;

    getblock(result.docid, tm, sz, url, author, title, digest);
    char *escaped_url = escape_url(url);
    char buf[blksize + 1];

    out << "<!-- id:" << result.docid << "-->" << endl;
    out << "<DT><B><A HREF=\"" << escaped_url << "\">";
    if(title[0]){
        out << html_entitize(buf, title);
    }else out << "(Untitled)";
    out << "</A></B>\n";
    out << " <SMALL>(score: " << (int)result.weight << ")</SMALL>\n";
    out << "<DD>";
    if(author[0]){
        out << "  Author: <TT>"
            << html_entitize(buf, author) << "</TT><BR>\n";
    }
    out << "&gt;&gt; " << html_entitize(buf, digest) << "..<BR>\n";
    out << "<NOBR><A HREF=\"" << escaped_url << "\">"
        << escaped_url << "</A></NOBR>";
    out << "<SMALL><NOBR>( " << sz << " bytes)</NOBR> <NOBR>";
    out << ctime(&tm) << "</NOBR></SMALL><BR><BR>\n";
    out << endl; 
    delete []escaped_url;
}

// ------------------------------------------------------------

void
MailDescription::print(ostream &out, const Result &result){
    char buf[blksize + 1];
    file.seekg(blksize * result.docid);
    file.read(buf, blksize);
    buf[blksize] = '\0';

    assert(sizeof(time_t) == sizeof(u_long));
    size_t sz = ntohl( *(size_t *)buf );
    time_t tm = ntohl( *(time_t *)(buf + sizeof(size_t)) );
    // time_t lm = ntohl( *(time_t *)(buf + 2) );
    char *path, *from, *subject, *digest;

    char *cp = buf + 2 * sizeof(time_t) + sizeof(size_t);
    path = cp;
    while(*cp) cp++;
    from = ++cp;
    while(*cp && cp < buf + blksize - 1) cp++;
    subject = ++cp;
    while(*cp && cp < buf + blksize - 1) cp++;
    digest = cp + 1;
    
    //    out << "* docid = " << result.docid << endl;
    out << '[' << path << ']';
    out << " (score: " << (int)result.weight;
    out << ", size: " << sz << " bytes)" << endl;
    out << "  Date: " << ctime(&tm);
    out << "  From: " << from << endl;
    out << "  Subject: " << subject << endl;
    out << ">> " << digest << ".." << endl;
    out << endl; 
}

