// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: HashBag.h,v 0.9 1997-12-27 07:14:00+09 harada Exp $
//
#ifndef HASHBAG_H_INCLUDED
#define HASHBAG_H_INCLUDED

#include<iostream.h>
#include<iomanip.h>
#include "HashKey.h"

#define PRIMETABLESIZE (256)

extern unsigned int primeTbl[];

template <class T>
class HashBag{
private:
    int slotNum;
    int usedNum;
    int cursor;
    HashKey **keyslot;
    T *valueslot;
    unsigned int hash2(unsigned int h){
        return primeTbl[h % PRIMETABLESIZE];
    }
public:
    HashBag(int n);
    HashBag(const HashBag &that);
    ~HashBag();
    int get(const HashKey &key, T &value);
    int set(const HashKey &key, const T &value);
    T & operator[](const HashKey &key);
    operator int() const { return usedNum; }
    int size() const { return usedNum; }
    int capacity() const { return slotNum; }
    void rewind(){ cursor = 0; };
    int next(HashKey *&keyp, T *&valuep);
    int printDebug();
};

// nʬΥǡ뤿Υϥåơ֥
// ưŪ˥礭ʤäϤʤ
template <class T>
HashBag<T>::HashBag(int n)
    : slotNum(16), usedNum(0), cursor(0)
{
    for(;;){
        if(slotNum > 3 * n) break;
        slotNum <<= 1;          // slotNum ɬ(äǿˤʤʤ)
    }
    keyslot = new (HashKey *)[slotNum];
    valueslot = new T[slotNum];
    memset(keyslot, 0, sizeof(HashKey *) * slotNum);
}


template <class T>
HashBag<T>::HashBag(const HashBag &that)
    : slotNum(that.slotNum), usedNum(that.usedNum), cursor(that.cursor)
{
    keyslot = new (HashKey *)[slotNum];
    valueslot = new T[slotNum];
    memset(keyslot, 0, sizeof(HashKey *) * slotNum);
    for(int i = 0; i < slotNum; i++){
        if(that.keyslot[i]){
            keyslot[i] = that.keyslot[i]->clone();
            valueslot[i] = that.valueslot[i];
        }
    }
}


template <class T>
HashBag<T>::~HashBag(){
    for(int i = 0; i < slotNum; i++)
        if(keyslot[i]) delete (keyslot[i]);
    if(keyslot) delete []keyslot;
    if(valueslot) delete []valueslot;
}


template <class T> int
HashBag<T>::printDebug(){
    cout << "---HashBag---" << endl;
    cout << "slotNum = " << slotNum << endl;
    cout << "usedNum = " << usedNum << endl;
    for(int i = 0; i < slotNum; i++){
        if(keyslot[i] != NULL){
            cout << setw(4) << i << ": \""
                 << keyslot[i]->toString() << "\" \t" << valueslot[i] << endl;
        }
    }
    return 0;
}


template <class T> int
HashBag<T>::set(const HashKey &key, const T &value){
    unsigned int h, h2;
    h = key.hashCode() % slotNum;
    h2 = hash2(h);
    while(NULL != keyslot[h]){
        if(key == *(keyslot[h])){
            valueslot[h] = value;
            return 0;
        }
        h = (h + h2) % slotNum;
    }
    keyslot[h] = key.clone();
    valueslot[h] = value;
    usedNum++;
    return 0;
}


template <class T> T &
HashBag<T>::operator[](const HashKey &key){
    unsigned int h, h2;
    h = key.hashCode() % slotNum;
    h2 = hash2(h);
    for(;;){
        if(NULL == keyslot[h]){
            usedNum++;
            keyslot[h] = key.clone();
            return valueslot[h];
        }
        if(key == *(keyslot[h])) return valueslot[h];
        h = (h + h2) % slotNum;
    }
}

// Ĥä0ʳ֤
template <class T> int
HashBag<T>::get(const HashKey &key, T &value){
    unsigned int h, h2;
    h = key.hashCode() % slotNum;
    h2 = hash2(h);
    for(;;){
        if(NULL == keyslot[h]) return 0;
        if(key == *(keyslot[h])){
            value = valueslot[h];
            return 1;
        }
        h = (h + h2) % slotNum;
    }
}

// ãƤ顤0 ֤
template <class T> int
HashBag<T>::next(HashKey *& keyp, T *&valuep){
    for( ; cursor < slotNum; cursor++){
        if(NULL != keyslot[cursor]){
            keyp = keyslot[cursor];
            valuep = &valueslot[cursor];
            cursor++;
            return 1;
        }
    }
    return 0;
}

#endif
