// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: IndexFile.cc,v 0.9 1997-12-27 07:17:06+09 harada Exp $
//

#include <iostream.h>
extern "C" {
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
}
#include "config.h"
#include "settemplate.h"
#include "IndexFile.h"

char IndexFile::MAGIC[] = "_FREYAIDXFILE_\n"; // 16 bytes

IndexFile::IndexFile(fstream &file, int rdonlyp, int newfilep)
    : idxfile(file), readonlyp(rdonlyp), blockNum(0)
{
    freeSizeTbl = new size_t[MAXBLOCKNUM];
    memset(freeSizeTbl, 0, MAXBLOCKNUM * sizeof(size_t));

    if(! newfilep){
        // read table from the file tail
        struct stat fstatus;
        fstat(idxfile.rdbuf()->fd(), &fstatus);
        off_t filesize = fstatus.st_size;
        // cerr << "filesize = " << filesize << endl;

        int tailblkNum; // Τtail tableȤƻȤƤ֥åο
        for(tailblkNum = 0; ; tailblkNum++)
            if((int)((filesize % BLOCKSIZE + tailblkNum * BLOCKSIZE
                      - sizeof(MAGIC))/sizeof(size_t))
               >= (filesize / BLOCKSIZE) - tailblkNum ) break;

        blockNum = filesize / BLOCKSIZE - tailblkNum;
        
        // check magic string
        char magic[sizeof(MAGIC)];
        idxfile.clear();
        idxfile.seekg(blockNum * BLOCKSIZE, ios::beg);
        idxfile.read(magic, sizeof(MAGIC));
        if(0 != strcmp(magic, MAGIC)){
            PRINTERRMSG("MAGIC check fail!");
        }else{
            // idxfile.seekg(0);
            idxfile.read(freeSizeTbl, sizeof(freeSizeTbl[0]) * blockNum);
            assert(sizeof(freeSizeTbl[0]) * blockNum ==
                   filesize - blockNum*BLOCKSIZE - sizeof(MAGIC));
        }
    }
}

IndexFile::~IndexFile(){
    if(! readonlyp){
        //cout << "write MAGIC, blockNum =" << blockNum << endl;
        idxfile.seekp(blockNum * BLOCKSIZE, ios::beg);
        idxfile.clear();
        idxfile.write(MAGIC, sizeof(MAGIC));
        idxfile.write(freeSizeTbl, blockNum * sizeof(freeSizeTbl[0]));
    }
    idxfile.close();
    if(freeSizeTbl) delete []freeSizeTbl;
}

void
IndexFile::printDebug() const {
    cout << "--- IndexFile::printDebug() ---" << endl;
    cout << "idxfile: good?=" << idxfile.good();
    cout << " eof?=" << idxfile.eof();
    cout << " fail?=" << idxfile.fail();
    cout << " bad?=" << idxfile.bad() << endl;
    cout << "readonlyp = " << readonlyp << endl;
    cout << "blockNum = " << blockNum << endl;
    if(! readonlyp){
        cout << "freeSizeTbl[] = ( ";
        for(int n = 0; n < blockNum; n++)
            cout << n << ":" << freeSizeTbl[n] << " ";
        cout << " )" << endl;
    }
    cout << "-------------------------------" << endl;
}

pos_t *
IndexFile::fetch(blockno_t bn, recno_t rn, int &listLen){
    listLen = 0;                // 顼ξ0ˤƤ
    if(TYPE_ERROR == getBlockType(bn)){
        PRINTERRMSG("internal error.");
        return NULL;
    }else{
        RecordBlock *recblock;
        size_t foundsize;
        char buf[BLOCKSIZE];

        readBlock(buf, bn);
        recblock = new RecordBlock(buf, BLOCKSIZE);
        char *clist = recblock->getRecord(rn, foundsize);
        //cout << "IndexFile::fetch bn = " << bn << flush;
        //cout << "  rn = " << rn << flush;
        if(NULL == clist){
            PRINTERRMSG("internal error.");
            recblock->printDebug();
            assert(clist != NULL);
            delete recblock;
            return NULL; // not found?
        }
        delete recblock;

        pos_t tmp[2 * foundsize + 1024];
        listLen = coder.diffunpack(clist, foundsize,
                                   tmp, sizeof(tmp)/sizeof(tmp[0]));
        if(listLen == (int)(sizeof(tmp)/sizeof(tmp[0]))){
            PRINTERRMSG("insufficient buffer - result maybe truncated.");
        }
        pos_t *list = new pos_t[listLen];
        memcpy(list, tmp, sizeof(pos_t) * listLen);
        delete []clist;

        return list;                //  delete
    }
}


// ĥ֥åɤ߹
pos_t *
IndexFile::fetchX(blockno_t bn, int &listLen){
    XRecordBlock *xrecblock;
    size_t blocksize = getBlockSize(bn);
    assert(blocksize > BLOCKSIZE);
    char buf[blocksize];
    readBlock(buf, bn, blocksize);
    xrecblock = new XRecordBlock(buf, blocksize);
    //cout << "fetchX: blocksize/BLOCKSIZE = " << blocksize/BLOCKSIZE << endl;
    //xrecblock->printDebug();
    listLen = 0;

    size_t csize;
    char *clist = xrecblock->getRecord(csize);
    delete xrecblock;
    if(NULL == clist){          // not found?
        assert(csize == 0);
        return NULL;
    }

    // compressing
    pos_t tmp[csize + 1024];
    listLen = coder.diffunpack(clist, csize, tmp, sizeof(tmp)/sizeof(tmp[0]));
    if(listLen == (int)(sizeof(tmp)/sizeof(tmp[0]))){
        PRINTERRMSG("insufficient buffer - result maybe truncated.");
    }
    pos_t *list = new pos_t[listLen];
    memcpy(list, tmp, sizeof(pos_t) * listLen);

    delete []clist;

    return list;                //  delete
}

// Ŭڤʥ֥åõ쥳ɤȤƳǼ
int
IndexFile::store(const pos_t *list, int listLen, 
                 blockno_t &newbn, recno_t &newrn, size_t &newsize){
    char packedlist[sizeof(pos_t) * listLen + 1024];
    size_t packedLen;
    if(NULL == coder.diffpack(list, listLen, packedlist,
                              sizeof(packedlist), packedLen)){
        PRINTERRMSG("buffer overrun(while compressing)");
    }

    newbn = findBlock(packedLen + sizeof(short_size_t) + sizeof(pos_t) + 1);
    if((blockno_t)-1 == newbn){
        // ĤʤäΤǡ֥åݤ
        return storeNewBlock(packedlist, packedLen, newbn, newrn, newsize);
    }
    // Ĥä¸Υ֥å(̾掠Υ֥å)ɲ
    char buf[BLOCKSIZE];
    readBlock(buf, newbn);
    RecordBlock *recblock = new RecordBlock(buf, BLOCKSIZE);

    if(RecordBlock::ERROVERFLOW
       == recblock->setRecord((recno_t)-1,packedlist,packedLen,newrn,newsize)){
        PRINTERRMSG("internal error.");
        cerr << " freesize = " << recblock->getFreeSize()
             << ", size = " << packedLen << endl;
        assert(recblock->getFreeSize() > (int)packedLen);
        delete recblock;
        return -1;              // 顼
    }
    updateFreeTbl(newbn, recblock->getFreeSize());
    delete recblock;
    writeBlock(buf, newbn, sizeof(buf));

    return 0;
}

// ֥åݤ˳Ǽ
int
IndexFile::storeNewBlock(const char *data, size_t datasize,
                         blockno_t &newbn, recno_t &newrn, size_t &newsize){
    if(datasize >= RecordBlock::MAXDATASIZE){
        newrn = (recno_t)-1;
        return storeNewBlockX(data, datasize, newbn, newsize);
    }else{
        // ̾ΥΥ֥å򿷤˺
        char buf[BLOCKSIZE];
        memset(buf, 0, sizeof(buf));
        RecordBlock *recblock = new RecordBlock(buf, BLOCKSIZE);

        newbn = blockNum;
        blockNum++;
        if(RecordBlockBase::ERROVERFLOW
           == recblock->setRecord((recno_t)-1, data, datasize,
                                  newrn, newsize)){
            PRINTERRMSG("internal error.");
            delete recblock;
            return -1;
        }
        updateFreeTbl(newbn, recblock->getFreeSize());
        delete recblock;
        writeBlock(buf, newbn, sizeof(buf));
    }
    return 0;
}

// ĥ֥åݤƳǼ
int
IndexFile::storeNewBlockX(const char *data, size_t datasize,
                          blockno_t &newbn, size_t &newsize){
    // ĥ֥å򿷤˺
    size_t blocksize = IndexFile::bestBlockSize(datasize);
    assert(blocksize >= 2 * BLOCKSIZE);
    char buf[blocksize];
    memset(buf, 0, sizeof(buf));
    XRecordBlock *xrecblock = new XRecordBlock(buf, blocksize);
    //xrecblock->printDebug();

    newbn = blockNum;
    //cout << "storeNewBlockX: newbn = " << newbn
    //     << ", blocksize = " << blocksize << endl;
    blockNum += blocksize / BLOCKSIZE;
    if(RecordBlockBase::ERROVERFLOW
       == xrecblock->setRecord(data, datasize, newsize)){
        PRINTERRMSG("internal error.");
        delete xrecblock;
        return -1;
    }
        
    updateFreeTbl(newbn, xrecblock->getFreeSize(), blocksize);
    delete xrecblock;
    writeBlock(buf, newbn, sizeof(buf));

    return 0;
}


// ¸Υ쥳ɤ˥ޡǳǼ
int
IndexFile::storeMerge(const pos_t *list, int listLen, int &newlistLen,
                      blockno_t bn, recno_t rn,
                      blockno_t &newbn, recno_t &newrn){
    RecordBlock *recblock;
    char buf[BLOCKSIZE];

    newbn = bn; newrn = rn;
    readBlock(buf, bn);
    recblock = new RecordBlock(buf, BLOCKSIZE);

    // ¸Υ쥳ɤɤ߹ǥޡ
    int tmpLen, mergedLen;

    pos_t *tmplist = fetch(bn, rn, tmpLen);
    pos_t mergedlist[tmpLen + listLen];
    mergedLen = mergeList(mergedlist, tmplist, tmpLen, list, listLen);
    delete []tmplist;
    newlistLen = mergedLen;

    // compressing
    size_t packedLen;
    char packed[sizeof(pos_t) * mergedLen + 1024];
    if(NULL == coder.diffpack(mergedlist, mergedLen, packed,
                              sizeof(packed), packedLen)){
        PRINTERRMSG("buffer overrun(while compressing)");
    }

    // ¸Υ쥳ɤ˥ޡꥹȤ񤭤
    size_t newsize;
    if(RecordBlock::ERROVERFLOW ==
       recblock->setRecord(rn, packed, packedLen, newrn, newsize)){
        // 줿Τǡ¸Υ쥳ɤϺơ
        recblock->removeRecord(rn);
        updateFreeTbl(bn, recblock->getFreeSize());
        delete recblock;
        writeBlock(buf, bn, sizeof(buf));

        // 򸫤ĤƳǼ롣
        return store(mergedlist, mergedLen, newbn, newrn, newsize);
    }
    updateFreeTbl(bn, recblock->getFreeSize());
    delete recblock;
    writeBlock(buf, bn, sizeof(buf));

    return 0;
}

// ¸γĥ֥å˥ޡƳǼ
int
IndexFile::storeXMerge(const pos_t *list, int listLen, int &newlistLen,
                       blockno_t bn, blockno_t &newbn){
    XRecordBlock *xrecblock;
    size_t blocksize = getBlockSize(bn);
    char buf[blocksize];

    newbn = bn;
    readBlock(buf, bn, blocksize);
    xrecblock = new XRecordBlock(buf, blocksize);

    // ¸Υ쥳ɤɤ߹ǥޡ
    int tmpLen, mergedLen;

    pos_t *tmplist = fetchX(bn, tmpLen);
    pos_t mergedlist[tmpLen + listLen];
    mergedLen = mergeList(mergedlist, tmplist, tmpLen, list, listLen);
    delete []tmplist;
    newlistLen = mergedLen;

    // compressing
    size_t packedLen;
    char packed[sizeof(pos_t) * mergedLen + 1024];
    if(NULL == coder.diffpack(mergedlist, mergedLen, packed,
                              sizeof(packed), packedLen)){
        PRINTERRMSG("buffer overrun(while compressing)");
    }

    // ¸Υ쥳ɤ˥ޡꥹȤ񤭤
    size_t newsize;
    if(RecordBlock::ERROVERFLOW ==
       xrecblock->setRecord(packed, packedLen, newsize)){
        // 줿Τǡ¸Υ쥳ɤϺơ
        xrecblock->removeRecord();
        updateFreeTbl(bn, xrecblock->getFreeSize(), blocksize);
        delete xrecblock;

        // 򸫤ĤƳǼ롣
        return storeNewBlockX(packed, packedLen, newbn, newsize);
    }
    updateFreeTbl(bn, xrecblock->getFreeSize(), blocksize);
    delete xrecblock;

    return 0;
}

// ------------------------------------------------------------
// private member functions

// åݤoverride
int
IndexFile::readBlock(char *buf, blockno_t bn, size_t size){
    idxfile.seekg(bn * BLOCKSIZE, ios::beg);
    idxfile.read(buf, size);
    if(idxfile.gcount() != size) return -1;
    return idxfile.rdstate();   // 顼ʤ 0
}

int
IndexFile::writeBlock(const char *buf, blockno_t bn, size_t size){
    idxfile.seekp(bn * BLOCKSIZE, ios::beg);
    idxfile.clear();
    idxfile.write(buf, size);
    return idxfile.rdstate();
}
        
// freesizeʾζ̤ĥ֥åõ
// ̾ΥΥ֥åǡζ̤ĤΤõ(worst fit)
blockno_t
IndexFile::findBlock(size_t freesize){
    blockno_t best = (blockno_t)-1;
    size_t bestSize = freesize + 1;
    for(blockno_t n = 0; n < blockNum; n++){
        if(freeSizeTbl[n] > bestSize && getBlockType(n) == TYPE_NORMAL){
            best = n;
            bestSize = freeSizeTbl[n];
        }
    }

    return best;
}

//
int
IndexFile::updateFreeTbl(blockno_t bn, size_t freesize, size_t blocksize){
    freeSizeTbl[bn] = freesize;
    if(blocksize > BLOCKSIZE){
        blockno_t n;
        for(n = 1; n < (blocksize / BLOCKSIZE); n++)
            freeSizeTbl[bn + n] = (size_t)-1;
        assert(bn <= (blockno_t)(blockNum - 1));
    }
    return 0;
}

int
IndexFile::getBlockType(blockno_t bn){
    if(bn >= blockNum) return TYPE_ERROR;
    if(freeSizeTbl[bn] == (size_t)-1)
        return TYPE_ERROR; // ĥ֥å
    if(bn < blockNum - 1 && freeSizeTbl[bn + 1] == (size_t)-1)
        return TYPE_XBLOCK;
    else return TYPE_NORMAL;
}

size_t
IndexFile::getBlockSize(blockno_t bn){
    if(bn >= blockNum) return (size_t)-1;
    if(freeSizeTbl[bn] == (size_t)-1) return (size_t)-1;
    size_t size = BLOCKSIZE;
    while(bn < blockNum - 1 && freeSizeTbl[bn + 1] == (size_t)-1){
        bn++;
        size += BLOCKSIZE;
    }
    return size;
}


#ifdef INDEXFILE_DEBUG
#include <fcntl.h>
// g++ -Wall -g -O2 -DINDEXFILE_DEBUG IndexFile.cc RecordBlock.o
int main(int argc, char *argv[]){
    int fd;
    int newfile = 1;
    const char *path = "default.idx";
    if(argc >= 2){
        newfile = 0;
        path = argv[1];
    }
//    fd = open(path, O_RDWR|O_CREAT|O_TRUNC, 0666);
    fd = open(path, O_RDWR|O_CREAT, 0666);

    IndexFile *indexfile = new IndexFile(fd, 0, newfile);

    blockno_t newbn = 0;
    recno_t newrn = 0;
    size_t newsize = 0;
    pos_t hoge[65536];
    for(int N = 1; N < 180; N++){
        for(int n = 1; n <= N*N; n++) hoge[n - 1] = n;
        if(-1 == indexfile->store(hoge, N*N, newbn, newrn, newsize)){
            cout << "Error!" << endl;
        }
        cout << " N*N = " << N*N;
        cout << " newbn = " << newbn;
        cout << " newrn = " << newrn;
        cout << " newsize = " << newsize << endl;
    }

//    indexfile->printDebug();
    cout << "***" << endl;

    int found;
    pos_t *result;
    if(newrn != (recno_t)-1)
        result = indexfile->fetch(newbn, newrn, found);
    else
        result = indexfile->fetchX(newbn, found);
    cout << found << " found... ";
    for(int i = found - 10; i < found; i++){
        cout << result[i] << ' ';
    }
    cout << endl;

    delete result;
    delete indexfile;

    close(fd);

    return 0;
}
#endif
