// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
//  inverted index (i.e. document reference file)
//  $Id: IndexFile.h,v 0.9 1997-12-27 07:17:35+09 harada Exp $
//
#ifndef INDEXFILE_H_INCLUDED
#define INDEXFILE_H_INCLUDED

#include <fstream.h>
extern "C" {
#include <stdlib.h>
}
#include "Coder.h"
#include "typedef.h"
#include "RecordBlock.h"

// size_t &newsize ɬפˤʤ̤ʤ⡣

class IndexFile {
public:
    enum { BLOCKSIZE = RecordBlock::BLOCKSIZE,
           MAXBLOCKNUM = 64 * 1024 };
    IndexFile(fstream &file, int rdonlyp, int newfilep = 0);
    virtual ~IndexFile();
    pos_t *fetch(blockno_t bn, recno_t rn, int &listLen);
    pos_t *fetchX(blockno_t bn, int &listLen);
    int store(const pos_t *list, int listLen,
              blockno_t &newbn, recno_t &newrn, size_t &newsize);
    int storeMerge(const pos_t *list, int listLen, int &newlistLen,
                   blockno_t bn, recno_t rn,
                   blockno_t &newbn, recno_t &newrn);
    int storeXMerge(const pos_t *list, int listLen, int &newlistLen,
                    blockno_t bn, blockno_t &newbn);
    void printDebug() const;
//    void printStat() const;
protected:
    enum { TYPE_ERROR = -1, TYPE_NORMAL = 0, TYPE_XBLOCK = 1 };
    static char MAGIC[];
    Coder coder;
    fstream &idxfile;
    const int readonlyp;
    size_t *freeSizeTbl;
    int blockNum;

    int storeNewBlock(const char *data, size_t datasize,
                      blockno_t &newbn, recno_t &newrn, size_t &newsize);
    int storeNewBlockX(const char *data, size_t datasize,
                       blockno_t &newbn, size_t &newsize);
    virtual int readBlock(char *buf, blockno_t bn,
                          size_t size = BLOCKSIZE);
    virtual int writeBlock(const char *buf, blockno_t bn,
                           size_t size = BLOCKSIZE);
    blockno_t findBlock(size_t freesize);
    int updateFreeTbl(blockno_t bn, size_t freesize,
                      size_t blocksize = BLOCKSIZE);
    int getBlockType(blockno_t bn);
    size_t getBlockSize(blockno_t bn);
    inline size_t bestBlockSize(size_t reqsize) const;
};

inline size_t
IndexFile::bestBlockSize(size_t reqsize) const{
    return ((reqsize * 3 / 2) / BLOCKSIZE + 1) * BLOCKSIZE;
}

#endif
