// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: LexiconEntry.cc,v 0.9 1997-12-27 07:19:16+09 harada Exp $
//
extern "C" {
#include <stdlib.h>
#include <string.h>
#include <assert.h>
}
#include "config.h"
#include "settemplate.h"
#include "LexiconEntry.h"
#include "IndexFile.h"

LexiconEntry::LexiconEntry()
    : key(NULL), bn((blockno_t)-1), rn((recno_t)-1), valueNum(0)
{
}

LexiconEntry::LexiconEntry(const LexiconEntry &that)
    : valueNum(that.valueNum)
{
    key = new char[strlen(that.key) + 1];
    strcpy(key, that.key);
    if(valueNum){
        values = new pos_t[valueNum];
        memcpy(values, that.values, valueNum * sizeof(pos_t));
    }else{
        bn = that.bn;
        rn = that.rn;
    }
}

LexiconEntry::LexiconEntry(const char *keystr, blockno_t b, recno_t r)
    : bn(b), rn(r), valueNum(0)
{
    key = new char[strlen(keystr) + 1];
    strcpy(key, keystr);
}

LexiconEntry::LexiconEntry(const char *keystr,
                           const pos_t *list, int listLen)
    : valueNum(listLen)
{
    assert(listLen <= 5);
    values = new pos_t[valueNum];
    for(int i = 0; i < valueNum; i++) values[i] = list[i];
    key = new char[strlen(keystr) + 1];
    strcpy(key, keystr);
}

LexiconEntry::~LexiconEntry(){
    if(key) delete []key;
    if(valueNum && values) delete []values;
}


void
LexiconEntry::printDebug(ostream &out) const{
    out << "-- LexiconEntry: key = \"" << key << "\"" << endl;
    out << "  valueNum = " << valueNum << endl;
    if(valueNum){
        out << "  values = ( ";
        for(int i = 0; i < valueNum; i++)
            out << values[i] << ' ';
        out << ')' << endl;
    }else
        out << "  bn = " << bn << ", rn = " << rn << endl;
}


LexiconEntry &
LexiconEntry::operator =(const LexiconEntry &that){
    if(valueNum && values) delete []values;

    valueNum = that.valueNum;
    if(valueNum){
        values = new pos_t[valueNum];
        for(int i = 0; i < valueNum; i++) values[i] = that.values[i];
    }else{
        bn = that.bn;
        rn = that.rn;
    }
    if(that.key == NULL){
        if(key) delete []key;
        key = NULL;
    }else{
        if(key == NULL || strlen(key) < strlen(that.key)){
            if(key) delete []key;
            key = new char[strlen(that.key) + 1];        
        }
        strcpy(key, that.key);
    }
    return *this;
}


pos_t *
LexiconEntry::fetch(IndexFile &indexfile, int &listLen) const{
    assert(key != NULL);
    if(valueNum){
        pos_t *r = new pos_t[valueNum];
        for(int i = 0; i < valueNum; i++) r[i] = values[i];
        listLen = valueNum;
        return r;
    }
    listLen = 0;
    if(bn == (blockno_t)-1)
        return NULL;
    if(rn == (recno_t)-1){
        return indexfile.fetchX(bn, listLen);
    }else{
        return indexfile.fetch(bn, rn, listLen);
    }
}

// ȥ𤷤ư֥ꥹȤ¸
LexiconEntry &
LexiconEntry::store(IndexFile &indexfile, const pos_t *list, int listLen){
    if((valueNum + listLen) <= 5){
        if(valueNum > 0){
            pos_t *mergelist = new pos_t[valueNum + listLen];
            valueNum = mergeList(mergelist, values, valueNum, list, listLen);
            delete []values;
            values = mergelist;
        }else{
            valueNum = listLen;
            values = new pos_t[valueNum];
            for(int i = 0; i < valueNum; i++) values[i] = list[i];
        }
        return *this;
    }

    size_t newsize;
    blockno_t newbn = (blockno_t)-1;
    recno_t newrn = (recno_t)-1;
    if(valueNum == 0 && bn != (blockno_t)-1){    // ¸Υ쥳ɤ˥ޡ
        int newlistLen;
        
        if(rn == (recno_t)-1){
            indexfile.storeXMerge(list, listLen, newlistLen, bn, newbn);
            if(newbn != (blockno_t)-1) bn = newbn;
        }else{
            indexfile.storeMerge(list, listLen, newlistLen,
                                 bn, rn, newbn, newrn);
            if(newbn != (blockno_t)-1) bn = newbn;
            if(newrn != (recno_t)-1)   rn = newrn;
        }
    }else{                      // IndexFile˥쥳ɤ
        if(valueNum){
            pos_t *tmplist = new pos_t[valueNum + listLen];
            valueNum = mergeList(tmplist, list, listLen, values, valueNum);
            indexfile.store(tmplist, valueNum, newbn, newrn, newsize);
            delete []tmplist;
            valueNum = 0;
        }else{
            assert(bn == (blockno_t)-1);
            indexfile.store(list, listLen, newbn, newrn, newsize);
        }
        bn = newbn;
        rn = newrn;
    }
    return *this;
}


ostream&
operator<<(ostream &stream, const LexiconEntry &entry){
    if(entry.valueNum){
        assert(entry.valueNum <= 5);
        stream << entry.key;
        stream << (char)(LexiconEntry::ENTITYENTRY1 - 1 + entry.valueNum);
        stream.write(entry.values, sizeof(pos_t) * entry.valueNum);
    }else{
        if(entry.bn == (blockno_t)-1){ // type A
            cerr << "Warning: null entry! " << entry.key << endl;
            stream << entry.key;
            stream << (char)LexiconEntry::NULLENTRY;
        }else if(entry.rn == (blockno_t)-1){ // type C
            //cerr << "output external : " << entry.key << endl;
            stream << entry.key;
            stream << (char)LexiconEntry::XREFERENCEENTRY;
            stream.write(&entry.bn, sizeof(entry.bn));
        }else{                  // type B
            stream << entry.key;
            stream << (char)LexiconEntry::REFERENCEENTRY;
            stream.write(&entry.bn, sizeof(entry.bn));
            stream.write(&entry.rn, sizeof(entry.rn));
        }
    }

    return stream;
}


istream&
operator>>(istream &stream, LexiconEntry &entry){
    char keybuf[MAXKEYLENGTH];
    char *dp;
    int entrytype = LexiconEntry::NULLENTRY;
    for(dp = keybuf; !stream.fail(); dp++){
        stream >> *dp;
        if(stream.eof()){
            LexiconEntry nullentry;
            entry = nullentry;
            return stream;
        }
        if(*(unsigned char *)dp < LexiconEntry::MAXENTRYTYPE){
            entrytype = *dp;
            *dp = '\0';
            break;
        }
    }

    if(entry.key) delete []entry.key;
    if(entry.valueNum) delete []entry.values;
    entry.key = new char[dp - keybuf + 1];
    strcpy(entry.key, keybuf);

    switch(entrytype){
    case LexiconEntry::NULLENTRY:
        cerr << "Warning: null entry?! " << endl;
        entry.bn = (blockno_t)-1;
        entry.rn = (recno_t)-1;
        entry.valueNum = 0;
        break;
    case LexiconEntry::XREFERENCEENTRY:
        stream.read(&entry.bn, sizeof(entry.bn));
        entry.rn = (recno_t)-1;
        entry.valueNum = 0;
        break;
    case LexiconEntry::REFERENCEENTRY:
        stream.read(&entry.bn, sizeof(entry.bn));
        stream.read(&entry.rn, sizeof(entry.rn));
        entry.valueNum = 0;
        break;
    case LexiconEntry::ENTITYENTRY1:
    case LexiconEntry::ENTITYENTRY2:
    case LexiconEntry::ENTITYENTRY3:
    case LexiconEntry::ENTITYENTRY4:
    case LexiconEntry::ENTITYENTRY5:
        entry.valueNum = entrytype - LexiconEntry::ENTITYENTRY1 + 1;
        entry.values = new pos_t[entry.valueNum];
        stream.read(entry.values, sizeof(pos_t) * entry.valueNum);
        break;
    default:
        PRINTERRMSG("Unexpected sequence found! (while reading Lexicon file)");
        cerr << "entrytype = " << entrytype << endl;
        break;
    }

    return stream;
}
