// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: LexiconFile.cc,v 0.9 1997-12-27 07:19:55+09 harada Exp $
//

// TODO: offsetTbl  Coder ưפ˰̤Ǥ

#include <iostream.h>
#include <iomanip.h>
extern "C" {
#include <assert.h>
#include <string.h>
}
#include "config.h"
#include "jlib.h"
#include "LexiconFile.h"

char LexiconFile::MAGIC[] = "_FREYALEXICON_\n"; // 16byte(including '\0')

LexiconFile::LexiconFile(fstream &f, int rdonly, int newfile)
    : file(f), readonly(rdonly)
{
    headerbuf = new char[HEADERSIZE];
    offsetTbl = new size_t[65536];
    if(newfile){
        // initialize header
        memset(headerbuf, 0, HEADERSIZE);
        memcpy(headerbuf, MAGIC, sizeof(MAGIC));
        assert(sizeof(size_t) == 4);
        memset(offsetTbl, 0, 65536 * sizeof(offsetTbl[0]));
        filesize = HEADERSIZE + 65536 * sizeof(offsetTbl[0]);
        writeHeader();
    }else{
        readHeader();
    }
}

LexiconFile::~LexiconFile(){
    if(! readonly){
        writeHeader();
    }
    file.close();
    delete []headerbuf;
    delete []offsetTbl;
}

void
LexiconFile::printDebug(ostream &out) const{
    out << "-- LexiconFile --" << endl;
    for(int i = 0; i < 256; i++){
        for(int j = 0; j < 256; j++){
            int k = i * 256 + j;
            if(offsetTbl[k]){
                out << "0x" << setfill('0') << setw(4) << hex << k << ' ';
                out << "0x" << offsetTbl[k] << " \""
                    << (unsigned char)i << (unsigned char)j << '"' << endl;
            }
            out << dec;
        }
    }
}

int
LexiconFile::readHeader(){
    file.seekg(0);
    file.clear();
    file.read(headerbuf, HEADERSIZE);
    if(file.gcount() != HEADERSIZE || file.fail()){
        PRINTERRMSG("Invalid lexicon file.");
        return 1;
    }
    if(0 != strcmp(headerbuf, MAGIC)){
        PRINTERRMSG("Invalid lexicon file! Please check the file.");
        return 2;
    }
    filesize = *(size_t *)(headerbuf + sizeof(MAGIC));
    file.read(offsetTbl, 65536 * sizeof(offsetTbl[0]));

    return file.fail();
}

int
LexiconFile::writeHeader(){
    file.seekp(0);
    file.clear();
    *(size_t *)(headerbuf + sizeof(MAGIC)) = filesize;
    file.write(headerbuf, HEADERSIZE);
    file.write(offsetTbl, 65536 * sizeof(offsetTbl[0]));
    return file.fail();
}

istream &
LexiconFile::rewind(){
    file.seekg(HEADERSIZE + sizeof(offsetTbl[0]) * 65536);
    return file;
}

void
LexiconFile::prepareAppend(){
    memset(offsetTbl, 0, 65536 * sizeof(offsetTbl[0]));
    file.seekp(HEADERSIZE + sizeof(offsetTbl[0]) * 65536);
    file.clear();
}

int
LexiconFile::indexhash(const char *key){
    unsigned char hi, lo;
    hi = *(unsigned char *)key++;
    lo = *(unsigned char *)key;
    return hi * 256 + lo;
}

int
LexiconFile::appendEntry(const LexiconEntry &entry){
    int idx = indexhash(entry.getkey());
    if(0 == offsetTbl[idx])
        offsetTbl[idx] = file.tellp();
    file << entry;
    return 0;
}


// exact match
int
LexiconFile::search(const char *key, LexiconEntry &result){
    size_t from, to;
    if(0 == findRegion(key, from, to))
        return 0;
    //cout << "LexiconFile::search: from = " << from << " to = " << to << endl;
    file.seekg(from);
    while(file.tellg() < (streampos)to){
        LexiconEntry entry;
        file >> entry;
        //cout << "  key = \"" << key << "\" entry = \""
        //     << entry.getkey() << "\"" << endl;
        int cmp = strcmp(entry.getkey(), key);
        if(cmp >= 0){
            if(0 == cmp){
                result = entry;
                return 1;       // found
            }else break;        // not found
        }
    }
    return 0;                   // not found
}

// ׸
//   : MAXPREFIXMATCHĤߤĤʤ
//         ɤȾ1ʸΤȤexact matchƱ
LexiconEntry **
LexiconFile::prefixsearch(const char *key, int &n){
    LexiconEntry **entries = new (LexiconEntry *)[MAXPREFIXMATCH + 1];
    size_t from, to;
    n = 0;
    if(0 == findRegion(key, from, to))
        return NULL;

    int keyLen = strlen(key);
    file.seekg(from);
    LexiconEntry *entryp;
    while(file.tellg() < (streampos)to){
        entryp = new LexiconEntry;
        file >> *entryp;
        int cmp = strncmp(entryp->getkey(), key, keyLen);
        if(cmp >= 0){
            if(cmp == 0){
                entries[n++] = entryp;
                if(n == MAXPREFIXMATCH) break;
            }else{
                delete entryp;
                break;
            }
        }else
            delete entryp;
    }
    entries[n] = NULL;
    return entries;
}

// find longest prefix (mainly for Japanese word breaking)
// : ѿʸ(1byte)prefixϸĤʤ
int
LexiconFile::findPrefix(const char *text, LexiconEntry &result){
    size_t from, to;
    if(0 == findRegion(text, from, to))
        return 0;               // not found
    //cout << "LexiconFile::findPrefix: from = " << from
    //     << " to = " << to << endl;
    file.seekg(from);
    int matchLen = 0;
    while(file.tellg() < (streampos)to){
        LexiconEntry entry;
        file >> entry;
        //cout << "  text = \"" << text << "\" entry = \""
        //     << entry.getkey() << "\"" << endl;
        int cmp = jisprefix(entry.getkey(), text);
        if(cmp >= 0){
            if(cmp == 0){
                matchLen = strlen(entry.getkey());
                result = entry;
            }else
                break;
        }
    }
    return matchLen;
}

int
LexiconFile::findRegion(const char *key, size_t &from, size_t &to){
    int idx = indexhash(key);
    from = offsetTbl[idx];
    if(0 == from)
        return 0;
    to = filesize;
    for(int i = idx + 1; i < 65536; i++){
        if(0 != offsetTbl[i]){
            to = offsetTbl[i];
            return 2;
        }
    }
    return 1;
}


#ifdef LEXICONFILE_DEBUG
// % g++ -O -g -Wall -DLEXICONFILE_DEBUG LexiconFile.cc LexiconEntry.cc IndexFile.cc RecordBlock.cc Coder.cc jlib.cc
int main(){
#if 0
    fstream lexfs("test.lex", ios::out|ios::in|ios::trunc, 0666);
    LexiconFile lex(lexfs, 0, 1);
    lex.printDebug();
#else
    fstream lexfs("test.lex", ios::in, 0666);
    LexiconFile lex(lexfs, 1, 0);
    istream &is = lex.rewind();
    for(;;){
        LexiconEntry entry;
        is >> entry;
        if(is.eof()) break;
        if(! entry.withValue() && entry.getrn() == (recno_t)-1){
            cout << "external! ";
            entry.printDebug();
        }
    }
    // lex.printDebug();
    
#endif
    return 0;
}
#endif
