// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: RecordBlock.cc,v 0.9 1997-12-27 07:21:50+09 harada Exp $
//
#include <iostream.h>
extern "C" {
#include <stdlib.h>
#include <string.h>
#include <assert.h>
}
#include "RecordBlock.h"

RecordBlockBase::RecordBlockBase(char *buf, size_t blocksize)
: buffer(buf), blockSize(blocksize)
{
}

// ------------------------------------------------------------

RecordBlock::RecordBlock(char *buf, size_t blocksize)
: RecordBlockBase(buf, blocksize),
  recordNum(*(unsigned short int *)buf),
  offsetTbl((short_size_t *)(buf + sizeof(recordNum)))
{
}

void
RecordBlock::printDebug(){
    cout << "--- RecordBlock::printDebug() ---" << endl;
    cout << "blocksize = " << getBlockSize() << endl;
    cout << "getFreeSize(): " << getFreeSize() << endl;
    cout << "recordNum = " << recordNum << endl;
    cout << "offsetTbl[] = { ";
    for(int i = 0; i < recordNum; i++)
        cout << offsetTbl[i] << ':' << getRecordLen(i) << ' ';
    cout << " }" << endl;
    cout << "---------------------------------" << endl;
}


int
RecordBlock::setRecord(recno_t rn, const char *list, size_t size,
                       recno_t &newrn, size_t &newsize){
    newrn = (recno_t)-1;
    newsize = 0;
    if(rn == (recno_t)-1){      // 쥳ɤ
        rn = findEmptyRecord();
        if(rn == (recno_t)-1){
            if((int)(size + sizeof(short_size_t)) >= getFreeSize())
                return ERROVERFLOW;

            if(recordNum == 0){      // ֥åˤǽΥ쥳ɤξ
                offsetTbl[recordNum] = (getBlockSize() - 1) - size;
            }
            else
                offsetTbl[recordNum] = offsetTbl[recordNum - 1] - size;
            memcpy(buffer + offsetTbl[recordNum], list, size);
            newrn = recordNum;
            newsize = size;
            recordNum++;
            
            return 0;
        }
        assert(getRecordLen(rn) == 0);
    }
    // ¸Υ쥳ɤ
    assert(rn < recordNum);
    int diff = size - getRecordLen(rn);
    if(diff >= getFreeSize()) return ERROVERFLOW;
    assert(diff > 0);
    if(rn < recordNum - 1){
        // ѹΤˤ餹
        assert(offsetTbl[recordNum - 1] - diff
               >= (int)(sizeof(recordNum) + sizeof(offsetTbl[0]) * recordNum));
        memmove(buffer + offsetTbl[recordNum - 1] - diff,
                buffer + offsetTbl[recordNum - 1],
                offsetTbl[rn] - offsetTbl[recordNum - 1]
                );
    }
    memcpy(buffer + offsetTbl[rn] - diff, list, size);
    for(int i = recordNum - 1; i >= rn; i--)
        offsetTbl[i] -= diff;
    newrn = rn;
    newsize = size;

    return 0;
}

char *
RecordBlock::getRecord(recno_t rn, size_t &recsize) const{
    if(rn >= recordNum) return NULL;

    int length = getRecordLen(rn);
    if(length <= 0){
        assert(length == 0);
        return NULL;
    }
    assert(length < 65536);
    char *result = new char[length];
    memcpy(result, buffer + offsetTbl[rn], length);
    recsize = length;
    
    return result;              // ꡼
}


int
RecordBlock::removeRecord(recno_t rn){
    // rn Ϸ֤ȤgetRecord()length0ˤʤ褦Ĵ
    if(rn >= recordNum) return -1;

    int length = getRecordLen(rn);
    if(length <= 0){             // Ǥ˺Ƥ?
        assert(length == 0);
        return -1;
    }
    // Τˤ餹
    if(rn < recordNum - 1){
        assert(offsetTbl[recordNum - 1] >= sizeof(recordNum));
        memmove(buffer + offsetTbl[recordNum - 1] + length,
                buffer + offsetTbl[recordNum - 1],
                offsetTbl[rn] - offsetTbl[recordNum - 1]
                );
    }
    memset(buffer + offsetTbl[recordNum - 1], 0, length);
    for(int i = recordNum - 1; i >= rn; i--)
        offsetTbl[i] += length;
    if(rn > 0)
        assert(offsetTbl[rn] == offsetTbl[rn - 1]);
    // ֤ˤʤΤǡrecordNum--; ɬפϤʤ
    // ǸΥ쥳ɤʤñ˺ǽ
    if(rn == recordNum - 1){
        offsetTbl[rn] = 0;
        recordNum--;
    }

    return 0;
}


recno_t
RecordBlock::findEmptyRecord(){
    for(recno_t n = 0; n < recordNum; n++)
        if(getRecordLen(n) == 0){
            return n;
        }
    return (recno_t)-1;
}

// ------------------------------------------------------------

XRecordBlock::XRecordBlock(char *buf, size_t blocksize)
: RecordBlockBase(buf, blocksize),
  recordSizep((size_t *)buf)
{
}

XRecordBlock::~XRecordBlock(){
}

void
XRecordBlock::printDebug(){
    cout << "--- XRecordBlock::printDebug() ---" << endl;
    cout << "blocksize = " << getBlockSize() << endl;
    cout << "getFreeSize(): " << getFreeSize() << endl;
    cout << "*recordSizep = " << *recordSizep << endl;
    cout << "----------------------------------" << endl;
}

int
XRecordBlock::setRecord(const char *list, size_t size, size_t &newsize){
    if((int)(size + sizeof(size_t)) >= getBlockSize()){
        newsize = 0;
        return ERROVERFLOW;
    }
    memcpy(buffer + sizeof(size_t), list, size);
    newsize = *recordSizep = size;
    return 0;
}

// 쥳ɤʣ֤ΨϤ褯ʤRecordBlock::getRecord()
// Ȼˡ碌ۤߥϸ
char *
XRecordBlock::getRecord(size_t &recsize) const{
    recsize = *recordSizep;
    char *result = new char[recsize];
    if(NULL == result){ recsize = 0; return NULL; }
    memcpy(result, buffer + sizeof(size_t), recsize);
    
    return result;              // ꡼
}

int
XRecordBlock::removeRecord(){
    memset(buffer, 0, getBlockSize());
    *recordSizep = 0;
    return 0;
}


#ifdef RECORDBLOCK_DEBUG
// % g++ -g -O2 -DRECORDBLOCK_DEBUG RecordBlock.cc
// % cat /usr/share/dict/words | ./a.out 
#include <unistd.h>
#include <fcntl.h>

int main(){
    char line[1024];
    char buffer[RecordBlock::BLOCKSIZE];
    int lineno = 0;

    memset(buffer, 0, sizeof(buffer));
    RecordBlock *rb = new RecordBlock(buffer, sizeof(buffer));
    recno_t newrn;
    size_t newsize;
    while(! cin.eof()){
        lineno++;
        cin >> line;
        if(line[0] == '\0') break;

        if(RecordBlock::ERROVERFLOW 
           == rb->setRecord((recno_t)-1, line, 1+strlen(line), newrn, newsize))
            {
                cout << "overflow!: " << rb->getFreeSize() << endl;
                break;
            }
    }

    cout << "before: " << rb->getRecordLen((recno_t)3) << endl;
    rb->removeRecord((recno_t)3); // 'aal'
    cout << "after: " << rb->getRecordLen((recno_t)3) << endl;
    if(RecordBlock::ERROVERFLOW
       == rb->setRecord((recno_t)-1, "bbb", 4, newrn, newsize)){
        cout << "overflow!:" << rb->getFreeSize() << endl;
    }

    int fd = open("test.blk", O_CREAT|O_RDWR|O_TRUNC, 0666);
    (void)lseek(fd, 0, SEEK_SET);
    write(fd, rb->getBuffer(), rb->getBlockSize());
    close(fd);

    delete rb;
    
    return 0;
}
#endif
