// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: RecordBlock.h,v 0.9 1997-12-27 07:22:26+09 harada Exp $
//
#ifndef RECORDBLOCK_H_INCLUDED
#define RECORDBLOCK_H_INCLUDED

#include "typedef.h"

class RecordBlockBase {
protected:
    char * const buffer;
    const int blockSize;
public:
    enum { SUCCESS = 0, ERROVERFLOW = 1 };
    enum { BLOCKSIZE = 64 * 1024 };
    RecordBlockBase(char *buf, size_t blocksize);
    virtual ~RecordBlockBase(){ }
    inline const char *getBuffer() const;
    inline int getBlockSize() const;
    virtual int getFreeSize() const = 0;
    virtual int getUsedSize() const = 0;
    virtual void printDebug() = 0;
};

inline const char *
RecordBlockBase::getBuffer() const { return buffer; }

inline int
RecordBlockBase::getBlockSize() const { return blockSize; }

// ------------------------------------------------------------

class RecordBlock : public RecordBlockBase {
private:
    unsigned short int &recordNum;
    short_size_t * const offsetTbl;
    recno_t findEmptyRecord();
public:
    enum { MAXDATASIZE = BLOCKSIZE - sizeof(short_size_t)
               - sizeof(unsigned short int) - 1};
    RecordBlock(char *buf, size_t blocksize = BLOCKSIZE);
    virtual ~RecordBlock(){ }
    int setRecord(recno_t rn, const char *list, size_t size,
                  recno_t &newrn, size_t &newsize);
    char *getRecord(recno_t rn, size_t &recsize) const;
    int removeRecord(recno_t rn);
    inline int getMaxDataSize() const;
    inline int getFreeSize() const;
    inline int getUsedSize() const;
    inline int getRecordLen(recno_t rn) const;
    void printDebug();
};

// ǼǽʺΥǡ
inline int
RecordBlock::getMaxDataSize() const {
    return blockSize - 1 - sizeof(short_size_t) - sizeof(recordNum);
}

inline int
RecordBlock::getFreeSize() const {
    if(recordNum == 0)
        return getMaxDataSize();
    assert(offsetTbl[recordNum - 1] >
           sizeof(recordNum) + recordNum * sizeof(short_size_t));
    return offsetTbl[recordNum - 1]
        - sizeof(short_size_t) * recordNum - sizeof(recordNum);
}

inline int
RecordBlock::getUsedSize() const {
    return getBlockSize() - 1 - getFreeSize();
}

inline int
RecordBlock::getRecordLen(recno_t rn) const {
    assert(rn < recordNum);
    if(rn == 0)
        return (getBlockSize() - 1) - offsetTbl[0];
    else
        return (offsetTbl[rn - 1] - offsetTbl[rn]);
}

// ------------------------------------------------------------

class XRecordBlock : public RecordBlockBase {
private:
    size_t * const recordSizep;
public:
    XRecordBlock(char *buf, size_t blocksize);
    ~XRecordBlock();
    int setRecord(const char *list, size_t size, size_t &newsize);
    char *getRecord(size_t &recsize) const;
    int removeRecord();
    inline int getMaxAllowedSize() const;
    inline int getFreeSize() const;
    inline int getUsedSize() const;
    void printDebug();
};

inline int
XRecordBlock::getMaxAllowedSize() const {
    return blockSize - sizeof(size_t);
}

inline int
XRecordBlock::getFreeSize() const{
    return getBlockSize() - getUsedSize();
}

inline int
XRecordBlock::getUsedSize() const{
    return sizeof(size_t) + *recordSizep;
}

#endif
