// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: WordBreaker.cc,v 0.9 1997-12-27 07:26:27+09 harada Exp $
//
extern "C" {
#include <assert.h>
#include <ctype.h>
}
#include "typedef.h"
#include "config.h"
#include "kcnvlib.h"
#include "Word.h"
#include "WordBreaker.h"
#include "valueSet.h"

WordBreaker::WordBreaker(Patricia &pat)
    : morphdic(pat)
{
}

WordBreaker::~WordBreaker(){
}

int
WordBreaker::is_valid_kanji(const char *kp) const{
    unsigned char valid[] = "";
    for(const unsigned char *vp = valid; *vp; vp += 2){
        if(       *vp == *(unsigned char *)kp
           && *(vp+1) == *(unsigned char *)(kp+1))
            return 1;
    }
    return Is_jis_level1(kp) || Is_jis_level2(kp)
        || Is_cyrillic(kp) || Is_greek(kp)
        || Is_hiragana_here(kp) || Is_katakana_here(kp);
}

int
WordBreaker::breakJapaneseText(HashBag<valueSet<pos_t> > &bag,
                               const char *text, pos_t &offset){
    // cout << "breakJapaneseText:\"" << text << "\"\n";
    int from = bag.size();
    Pat_node *prefixes[MAXPREFIXES];
    const char *cp = text;
    const char *cpforward = cp; // ϿѤߤʬ
    while(*cp && *(cp + 1)){
        int matchNum = morphdic.search_prefix(cp, prefixes);
        if(matchNum == 0
           || strlen(prefixes[matchNum - 1]->key) <= 4 ){ // 4 ˤ٤?
            if(*(cp + 2) == '\0'){ // Ǹΰʸ
                if(cpforward < cp + 2){
                    Word term(cp, 2);
                    bag[term] += offset;
                    cpforward = cp + 2;
                }
            }else{              // 2ʸ(4byte)Ͽ1ʸʤ
                Word term(cp, 4);
                bag[term] += offset;
                cpforward = cp + 4;
            }
            offset += 2;
            cp += 2;
        }else{
            // 3ʸ(6byte)ʾĹ줬ޥå
            Word term = prefixes[matchNum - 1]->key;
            if(term.length() <= MAXKEYLENGTH - 1) // -1 ǰΤ
                bag[term] += offset;
            cp += term.length();
            cpforward = cp;
            offset += term.length();
        }
    }
    return bag.size() - from;
}


int
WordBreaker::entryWord(HashBag<valueSet<pos_t> > &bag, const char *word,
                       int wordLen, pos_t offset, int method){
    char buf[wordLen + 1];
    int allcapital = 1;

    const char *cp = word;
    char *dp = buf;
    for(int i = 0; i < wordLen; i++){
        if(islower(*cp)){
            allcapital = 0;
            *dp++ = *cp++;	// just a copy
        }else{
            if(isupper(*cp))
                *dp++ = *cp++ + 'a' - 'A'; // decapitalize
            else
                *dp++ = *cp++;	// just a copy
        }
    }
    *dp = '\0';

    if(method == FORINDEX){
        // ʸΤϿ
        // ٤ʸʤ餽Ͽ
        Word term(buf, wordLen);
        bag[term] += offset;
        if(allcapital && 0 != strcmp(word, buf)){
            Word origterm(word, wordLen);
            bag[origterm] += offset;
        }
    }else{
        // ٤ʸʤ餽Ͽ
        // ǤʤȤϾʸΤϿ
        if(allcapital){
            Word origterm(word, wordLen);
            bag[origterm] += offset;
        }else{
            Word term(buf, wordLen);
            bag[term] += offset;
        }
    }
    return allcapital;
}


int
WordBreaker::breakText(HashBag<valueSet<pos_t> > &bag, const char *text,
                       pos_t &offset, int method){
    char buf[MAXSTRINGLENGTH + 1];
    char *kp;
    int from = bag.size();
    
    for(const char *cp = text; *cp; ){
        if(Is_kanji_here(cp)){
            if(! is_valid_kanji(cp)){
                cp += 2;
                offset++;
            }else{
                kp = buf;
                do {
                    *kp++ = *cp++; *kp++ = *cp++;
                }while(Is_kanji_here(cp) && is_valid_kanji(cp));
                *kp = '\0';
                (void) breakJapaneseText(bag, buf, offset);
            }
        }else if(is_valid_head(*cp)){	// ñ
			int bufLen = 0;
			char *bufp = buf;
			do {
                *bufp++ = *cp++;
                bufLen++;
            } while(is_valid_char(*cp));
			*bufp-- = '\0';
			// Ŭڤʸ ('.' ʤ)
			while(bufp > buf && (! is_valid_tail(*bufp)))
				*bufp-- = '\0';
            int termLen = bufp + 1 - buf;
            
            assert(termLen > 0);
            if(termLen <= MAXKEYLENGTH - 1)
                (void) entryWord(bag, buf, termLen, offset, method);
			offset += bufLen;
		}else if(Ishkana1(*cp) && Ishkana2(*(cp + 1))){
            // ȾѥʤskipbreakText()Ƥ
			cp += 2;
			offset += 2;
		}else{					// ¾skip
			cp++; offset++;
			while(isspace(*cp)) cp++; // ϤޤȤ skip Ƥ
		}
    }
    
    return bag.size() - from;
}

#if WORDBREAKER_DEBUG
// % g++ -O -g -DWORDBREAKER_DEBUG WordBreaker.cc patricia.cc HashBag.cc Word.cc
#include <fstream.h>
int main(){
    fstream morphdicfile("test.dic", ios::in);
    Patricia morphdic(morphdicfile);
    morphdicfile.close();
    if(! morphdic){
        cerr << "abort.\n";
        return 1;
    }

    HashBag<valueSet<pos_t> > words(128);
    WordBreaker wb(morphdic);
    pos_t offset = 0;
    wb.breakText(words, "Ϥmasa&no,ri that. 3com that.is C++ +p 123 harada@graco.c.u-tokyo.ac.jp ri,that", offset, WordBreaker::FORINDEX);
    offset = 1000;
    wb.breakText(words, "òɤ", offset, WordBreaker::FORINDEX);
    words.printDebug();

    return 0;
}
#endif
