// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: fsearch.cc,v 0.10 1997-12-29 14:23:20+09 harada Exp $
//
#include <iostream.h>
#include <strstream.h>
extern "C" {
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
}
#include "Retriever.h"
#include "Unifier.h"
#include "Expression.h"
#include "Evaluator.h"
#include "Result.h"
#include "Description.h"

#ifdef FSEARCHCGI
#include "kcnvlib.h"
#include "htmlutil.h"

// replace keyword("$foo") as str
int
replace_string(char *buf, const char *keyword, const char *str){
    char *cp;
    int keyLen = strlen(keyword);
    if((cp = strstr(buf, keyword))){
        memmove(cp + strlen(str), cp + keyLen,
                strlen(buf) - (cp + keyLen - buf) + 1);
        strncpy(cp, str, strlen(str));
        return 1;               // replaced
    }
    return 0;
}

int
display_html(ostream &out, const char *filepath,
             const char *title, const char *query, const char *errormsg,
             const char *index, int resultNum){
    fstream htmlfile(filepath, ios::in);
    if(! htmlfile.good())
        return -1;

    char qtitle[LINE_MAX];
    char qerrormsg[LINE_MAX];
    char qquery[LINE_MAX];
    html_entitize(qtitle, title);
    html_entitize(qerrormsg, errormsg);
    html_entitize(qquery, query);
    ostrstream numstr;
    numstr << ' ' << resultNum << '\0';
    char *num = numstr.str();

    char linebuf[LINE_MAX];
    for(;;){
        htmlfile.getline(linebuf, LINE_MAX, '\n');
        if(htmlfile.fail()){
            if(htmlfile.eof()) break; else cerr << "error.\n";
        }

        for(;;){
            int replaced = 0;
            replaced += replace_string(linebuf, "$title", qtitle);
            replaced += replace_string(linebuf, "$errormsg", qerrormsg);
            replaced += replace_string(linebuf, "$query", qquery);
            replaced += replace_string(linebuf, "$index", index);
            replaced += replace_string(linebuf, "$result", num);
            if(replaced == 0){
                out << linebuf << endl;
                break;
            }
        }
    }
    htmlfile.close();
    delete []num;
    return 0;
}

#if 0
void
cgi_error(const char *msg = NULL){
    cout << "Content-type: text/plain\n\n";
    if(msg) cout << msg << endl;
}
#else
void
cgi_error(const char *msg1, const char *msg2 = NULL,
          const char *query = NULL, const char *index = NULL){
    if(index == NULL) index = "";
    if(query == NULL) query = "";
    if(msg2 == NULL) msg2 = "";
    cout << "Content-type: text/html\n\n";
    display_html(cout, "error.html", msg1, query, msg2, index, 0);
}

#endif

int
get_query(char *key, char *index, int &from, int &dispNum){
    char *method = getenv("REQUEST_METHOD");
    if(! (method && 0 == strcmp(method, "GET"))) {
        if(method){
            cgi_error("Error", "unsupported method");
        }else{
            cgi_error("Error", "no method provided");
        }
        return 1;
    }
    char *qs = getenv("QUERY_STRING");
    if(NULL == qs){
        cgi_error("Error", "No query information to decode.");
        return 1;
    }
    char *k, *v;
    while((k = strtok(qs, "&"))){
        qs = NULL;
        v = strchr(k, '=');
        if(v != NULL){
            *v++ = '\0';
            unescape_url(v);
            // cout << "k = \"" << k << "\", param = \"" << param << "\"\n";
            if(0 == strcmp(k, "key")){
                if(NULL == conv2euc(key, v, CODE_UNKNOWN)){
                    strncpy(key, v, MAXQUERYLENGTH);
                }
            }else if(0 == strcmp(k, "from")){
                from = atoi(v);
            }else if(0 == strcmp(k, "index")){
                strncpy(index, v, PATH_MAX);
            }else if(0 == strcmp(k, "n")){
                dispNum = atoi(v);
            }
        }
    }

    return 0;                   // success
}

int
print_results(HtmlDescription *desc, const Result *result, const char *key,
              const char *index, int resultNum, int disp, int from, int to){
    assert(to >= from);
    assert(to - from < resultNum);
    int r;
    cout << "<DL>\n";
    for(r = from; r <= to; r++){
        desc->printHtml(cout, result[r]);
    }
    cout << "</DL>" << endl;
    
    char *cgipath = getenv("SCRIPT_NAME");
    if(NULL == cgipath) cgipath = "fsearch.cgi";
    char *escaped_key = escape_url(key);

    cout << "\n<BR>\n<BR>\n<SMALL><NOBR>\n";
    if(from > 0){
        cout << "<A HREF=\"" << cgipath << "?key=" << escaped_key
             << "&from=" << ((from - disp < 0) ? 0 : from - disp);
        cout << "&n=" << disp << "&index=" << index  << "\">"
             << "[PREV]</A>" << endl;
    }
    int i, p;
    int i_from = from - 10 * disp;
    if(i_from < 0) i_from = 0;
    int i_to = i_from + 20 * disp;
    if(i_to > resultNum) i_to = resultNum;

    for(i = i_from,p = i / disp; i < i_to; i += disp, p++){
        if(i == from)
            cout << "<B>[" << p << "]</B>" << endl;
        else{
            cout << "<A HREF=\"" << cgipath << "?key=" << escaped_key
                 << "&from=" << i << "&n=" << disp
                 << "&index=" << index << "\">"
                 << '[' << p << "]</A>" << endl;
        }
    }
    if(to + 1 < resultNum){
        cout << "<A HREF=\"" << cgipath << "?key=" << escaped_key
             << "&from=" << to + 1 << "&n=" << disp
             << "&index=" << index << "\">"
             << "[NEXT]</A>" << endl;
    }
    cout << "</SMALL></NOBR>\n";
    delete []escaped_key;

    return 0;
}

int
main(){
    char query[MAXQUERYLENGTH + 1];
    char indexbase[PATH_MAX + 1];
    int rfrom;
    int dispNum = RESULTSFORDISPLAY;

    query[0] = indexbase[0] = '\0';
    if(0 != get_query(query, indexbase, rfrom, dispNum)){ // error
        return 0;
    }
    rfrom++;
    if(indexbase[0] == '\0'){
        // cgi_error("Error", "No index specified.");
        strcpy(indexbase, "default");
    }
    if(query[0] == '\0'){
        cgi_error("顼: Ĥޤ",
                  "ϤƤ", "", indexbase);
        return 0;
    }
    if(dispNum < 0 || dispNum > MAXRESULTSFORDISPLAY){
        cgi_error("顼: ϰϤǤ",
                  "n=...ŬڤͤˤƤ", "", indexbase);
        return 0;
    }

    // setup files ...
    char indexpath[PATH_MAX];
    char lexpath[PATH_MAX];
    char xmappath[PATH_MAX];
    strcpy(indexpath, indexbase); strcat(indexpath, ".idx");
    strcpy(lexpath, indexbase); strcat(lexpath, ".lex");
    strcpy(xmappath, indexbase); strcat(xmappath, ".map");

    fstream idxstream(indexpath, ios::in);
    if(! idxstream.good()){
        cgi_error("Error", "cannot open index file.", query, indexbase);
        return 1;
    }
    IndexFile idxfile(idxstream, 1, 0); // readonly, !newfile
    fstream lexfs(lexpath, ios::in);
    if(! lexfs.good()){
        cgi_error("Error", "cannot open lexicon file.", query, indexbase);
        return 1;
    }
    LexiconFile lexfile(lexfs, 1, 0); // readonly, !newfile

    int fd = open(xmappath, O_RDONLY, 0666);
    if(fd == -1){
        cgi_error("Error", "cannot open map file", query, indexbase);
        return 1;
    }

    // setup retriever object
    XMap map(fd);
    Retriever retr(lexfile, idxfile, map);
    Evaluator evaluator;           // use standard ranking policy

    char uquery[MAXQUERYLENGTH + 1];
    const Unifier unif;
    unif.unify(uquery, query);
    ComplexExpression cexp(uquery);
    if(cexp.error()){
        cgi_error("顼: Ǥ",
                  cexp.errmsgs[cexp.error()], uquery, indexbase);
        return 2;
    }

    // do search!
    int resultLen = 0;
    Result *result = cexp.eval(retr, evaluator, resultLen);
    if(resultLen <= 0){
        cgi_error("ĤޤǤ", "nothing found. ",
                  uquery, indexbase);
    }else{
        evaluator.rank(result, resultLen);
        char descpath[PATH_MAX];
        strcpy(descpath, indexbase); strcat(descpath, ".dsc");
        fstream descfile(descpath, ios::in);
        if(! descfile.good()){
            cgi_error("Error", "cannot open description file.",
                      uquery, indexbase);
            return 3;
        }

        if(rfrom <= 0 || rfrom > resultLen){
            cgi_error("Error", "parameter 'from' out of range.",
                      uquery, indexbase);
            return 4;
        }
        int rto = rfrom + dispNum - 1;
        if(rto > resultLen) rto = resultLen;
        if(rto == 0)
            rto = resultLen;
        else
            rto = (resultLen > rto) ? rto : resultLen;

        ostrstream stitle, squery;
        stitle << "Result: ";
        cexp.print(stitle);
        stitle << " (" << resultLen << " documents found. #" << rfrom;
        stitle << "--#" << rto << ')' << '\0'; // terminate explicitly
        cexp.print(squery); squery << '\0';
        cout << "Content-type: text/html\n\n";
        char *titlestr = stitle.str();
        char *querystr = squery.str();
        display_html(cout, "head.html", titlestr, querystr,
                     "(no error)", indexbase, resultLen);
        HtmlDescription *desc = new HtmlDescription(descfile);
        print_results(desc, result, uquery, indexbase,
                      resultLen, dispNum, rfrom -1, rto - 1);
        delete desc;
        display_html(cout, "foot.html", titlestr, querystr,
                     "(no error)", indexbase, resultLen);
        delete []titlestr;
        delete []querystr;
    }
    if(result) delete []result;
    
    return 0;
}

#else // FSEARCHCGI

void
printusage(){
    cout << FREYABANNER << endl
         << "Usage: fsearch [-s mail|html] [-f n] [-t n] -i index key [key ..]" << endl;
}

char *
mystrdup(const char *str){
    int l = strlen(str);
    char *s = new char[l + 1];
    strcpy(s, str);
    return s;
}

void
buildquery(char *q, int argc, char *argv[]){
    static Unifier unif;
    char buf[MAXKEYLENGTH + 1];
    char unified[MAXKEYLENGTH + 16 + 1];
    char *p = q;
    for(int i = 0; i < argc; i++){
        strncpy(buf, argv[i], MAXKEYLENGTH);
        unif.unify(unified, buf);
        char *cp = unified;
        while((*p = *cp++) && p - q < MAXQUERYLENGTH)
            p++;
        *p++ = ' ';
    }
    if(*p == ' ' && p > q) p--;
    *p = '\0';
}

int
main(int argc, char *argv[]){
    char *indexbase = "default";
    char *style = NULL;
    extern char *optarg;
    extern int optind; // opterr, optopt;
    int rfrom = 1;
    int rto = RESULTSFORDISPLAY;

    int ch;
    while((ch = getopt(argc, argv, "f:t:s:i:v?")) != EOF){
        switch(ch){
        case 's':
            style = mystrdup(optarg);
            break;
        case 'i':
            indexbase = mystrdup(optarg);
            break;
        case 'f':
            rfrom = atoi(optarg);
            break;
        case 't':
            rto = atoi(optarg);
            break;
        case 'v':
        case '?':
        default:
            printusage();
            return 1;
        }
    }
    argc -= optind;
    argv += optind;

    if(! argv[0]){
        printusage();
        return 1;
    }
    if(NULL == style){
        style = "mail";
    }else if(0 != strcmp(style, "mail") && 0 != strcmp(style, "html")){
        PRINTERRMSG("unknown style(-s) specified.");
        cerr << "  style = \"" << style << '"' << endl;
        return 1;
    }

    // rto == 0 stands for `all result'
    if((rto > 0 && rto < rfrom) || rfrom <= 0 || rto < 0){
        PRINTERRMSG("inappropriate range specified.");
        printusage();
        return 1;
    }

    char query[MAXQUERYLENGTH + 1];
    buildquery(query, argc, argv);

    // setup files ...
    char indexpath[PATH_MAX];
    char lexpath[PATH_MAX];
    char xmappath[PATH_MAX];
    strcpy(indexpath, indexbase); strcat(indexpath, ".idx");
    strcpy(lexpath, indexbase); strcat(lexpath, ".lex");
    strcpy(xmappath, indexbase); strcat(xmappath, ".map");

    fstream idxstream(indexpath, ios::in);
    if(! idxstream.good()){
        PRINTERRMSG("cannot open index file.");
        cerr << "  indexpath = \"" << indexpath << "\"\n";
        return 1;
    }
    IndexFile idxfile(idxstream, 1, 0); // readonly, !newfile
    fstream lexfs(lexpath, ios::in);
    if(! lexfs.good()){
        PRINTERRMSG("cannot open lexicon file.");
        cerr << "  lexpath = \"" << lexpath << "\"\n";
        return 1;
    }
    LexiconFile lexfile(lexfs, 1, 0); // readonly, !newfile

    int fd = open(xmappath, O_RDONLY, 0666);
    if(fd == -1){
        PRINTERRMSG("cannot open map file");
        cerr << "  xmappath = \"" << xmappath << "\"\n";
        return 1;
    }

    // setup retriever object
    XMap map(fd);
    Retriever retr(lexfile, idxfile, map);
    Evaluator evaluator;        // use standard ranking policy

    ComplexExpression cexp(query);
    if(cexp.error()){
        cout << "Error: " << cexp.errmsgs[cexp.error()] << endl;
        return 2;
    }

    // do search!
    int resultLen = 0;
    Result *result = cexp.eval(retr, evaluator, resultLen);
    if(resultLen > 0){
        evaluator.rank(result, resultLen);

        char descpath[PATH_MAX];
        strcpy(descpath, indexbase); strcat(descpath, ".dsc");
        fstream descfile(descpath, ios::in);
        if(! descfile.good()){
            PRINTERRMSG("cannot open description file.");
            cerr << " descpath = \"" << descpath << "\"\n";
            return 3;
        }

        Description *desc;
        if(0 == strcmp(style, "mail"))
            desc = new MailDescription(descfile);
        else
            desc = new HtmlDescription(descfile);
        cout << "Result: " << resultLen << " documents found.:" << endl;
        if(rto == 0)
            rto = resultLen;
        else
            rto = (resultLen > rto) ? rto : resultLen;
        int r;
        for(r = 0; r < resultLen; r++){
            cout << result[r].docid << ' ';
        }
        cout << endl;
        for(r = rfrom - 1; r <= rto - 1; r++){
            desc->print(cout, result[r]);
        }
        cout << endl;
        delete desc;
    }else{
        cout << "nothing found.\n";
        cexp.printDebug();
    }
    if(result) delete []result;
    // cexp.printDebug();

    return 0;
}
#endif // FSEARCHCGI
