// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// Patricia class
// $Id: patricia.cc,v 0.9 1997-12-27 07:50:38+09 harada Exp $
//
extern "C" {
#include <assert.h>
#include <string.h>
}
#include "patricia.h"

#ifdef PATRICIA_DEBUG
static int allocated_node = 0;
static int released_node = 0;
#endif

// ----------------------------------------

Pat_node::Pat_node()
    : key(NULL), content(0), checkbit(0), left(NULL), right(NULL)
{
#ifdef PATRICIA_DEBUG
    allocated_node++;
#endif
    // -1 ƤơŤֹ꿶ʤ褦ˤ롣(Ad Hoc)
    dump_idx = -1;
}

Pat_node::Pat_node(const char *k, long c)
    : content(c), checkbit(0), left(NULL), right(NULL)
{
#ifdef PATRICIA_DEBUG
    allocated_node++;
#endif
    int len = strlen(k);
    key = new char[len + 1];
    strcpy(key, k);
    // -1 ƤơŤֹ꿶ʤ褦ˤ롣(Ad Hoc)
    dump_idx = -1;
}


Pat_node::~Pat_node(){
#ifdef PATRICIA_DEBUG
    released_node++;
#endif
    if(key) delete []key;
}

// ----------------------------------------

Patricia::Patricia() : root_node(NULL), prefix_dump_counter(0), keyNum(0) {
}

Patricia::Patricia(istream &file)
     : root_node(NULL), prefix_dump_counter(0), keyNum(0)
{
    load(file);
}

Patricia::~Patricia(){
    if(NULL == root_node) return;
    delete_subtree(root_node);
    root_node = NULL;           // ǰΤ
}


// ƵŪʬڤ
void
Patricia::delete_subtree(Pat_node *here){
    if(here->left->checkbit > here->checkbit) delete_subtree(here->left);
    if(here->right->checkbit > here->checkbit) delete_subtree(here->right);
    delete here;
}

int
Patricia::size(){
    return keyNum;
}


// bitܤۤʤäƤ뤫Ĵ٤롣ۤʤäƤʤ -1 ֤
// char  8bit Ǥ뤳Ȥ()
int
Patricia::find_diffbit(const char *s1, const char *s2){
    for(int nth = 0; ; nth++){
        if(s1[nth] != s2[nth]){
            int c1 = s1[nth];
            int c2 = s2[nth];
            int mask = 0x80;
            for(int nthbit = 0; ; nthbit++){
                if((c1 & mask) != (c2 & mask)) return nth * 8 + nthbit;
                mask >>= 1;
                assert(mask);
                assert(nthbit < 8);
            }
        }else 
            if('\0' == s1[nth]) return -1;
    }
}


int
Patricia::getbit(const char *string, int nthbit, int len){
    if(nthbit == -1) nthbit = 0; // root_node к
    int n = nthbit / 8;
    if(n >= len){ return 0; } // ʸαˤ 0 ߤͤƤȲ
    char c = string[n];
    nthbit = nthbit % 8;
    return (c >> (8 - nthbit - 1)) & 0x01;
}


Pat_node *
Patricia::down_node(const char *key, Pat_node *root){
    Pat_node *forw, *here;
    forw = root;
    do{
        here = forw;
        forw = getbit(key, here->checkbit) ? here->right : here->left;
    }while(here->checkbit < forw->checkbit);
    return forw;
}


// -1 : Ĥʤä
long
Patricia::search(const char *key){
    Pat_node *bottom;
    if(NULL == root_node) return -1;
    bottom = down_node(key, root_node); // ꤽʥΡɤޤǹߤ
    if(0 == strcmp(key, bottom->key)){
//      cerr << "found: " << key << ", " << bottom->content << endl;
        return bottom->content;
    }
    return -1;
}


// s1  s2  prefix ɤĴ٤
// : NULL, : s2  s1 ʹߤʬؤΥݥ
const char *
Patricia::is_prefix(const char *s1, const char *s2){
    while(*s1 && (*s1 == *s2)){
        s1++; s2++;
    }
    if('\0' == *s1) return s2; else return NULL;
}


// key  prefix ȤƻĤ褦ʸõ (see also search_prefix())
int
Patricia::prefix_search(const char *key, Pat_node *matches[]){
    Pat_node *prev, *here, *forw;
    int keylen;

    if(NULL == root_node) return -1;
    keylen = strlen(key);
    prefix_dump_counter = 0;
    here = forw = root_node;
    do{                         // ʬޤǹߤ
        prev = here;
        here = forw;
        if(here->checkbit >= keylen * 8)    // ĤϤŤõ
            return prefix_dump(key, matches, prev);
        forw = getbit(key, here->checkbit) ? here->right : here->left;
    }while(here->checkbit < forw->checkbit);

    if(is_prefix(key, forw->key))   // Ǹ˰Ĥ뤫
        matches[prefix_dump_counter++] = forw;
    return prefix_dump_counter;
}


// ƵŪʬƬõΤˤ
// ʬڤˤΡɤ˥󥯤ƤΥΡɤõ
int
Patricia::prefix_dump(const char *key, Pat_node *matches[], Pat_node *here)
{
    if(here->left->checkbit > here->checkbit) 
        prefix_dump(key, matches, here->left);
    else
        if(is_prefix(key, here->left->key))
            matches[prefix_dump_counter++] = here->left;

    if(here->right->checkbit > here->checkbit)
        prefix_dump(key, matches, here->right);
    else 
        if(is_prefix(key, here->right->key))
            matches[prefix_dump_counter++] = here->right;

    return prefix_dump_counter;
}

int
Patricia::load(istream &patfile){
    patfile.seekg(0);

    // read header
    unsigned int magic;
    int root_idx;
    patfile.read(&magic, sizeof(magic));
    if(magic != MAGIC){
        cerr << "Patricia::load(): invalid PAT file." << endl;
        return -1;
    }
    patfile.read(&keyNum, sizeof(keyNum));
    patfile.read(&root_idx, sizeof(root_idx));

    // prepare body
    Pat_node *node_table[keyNum];
    int i;
    for(i = 0; i < keyNum; i++)
        node_table[i] = new Pat_node();

    // read body
    for(i = 0; i < keyNum; i++){
        // read key
        Pat_node *node = node_table[i];
        unsigned char keylen;
        patfile.get(keylen);    // may keylen be 0
        node->key = new char[keylen + 1];
        patfile.read(node->key, (int)keylen);
        node->key[keylen] = '\0';

        // read content, checkbit
        patfile.read(& node->content, sizeof(node->content));
        patfile.read(& node->checkbit, sizeof(node->checkbit));

        // read branches
        int left_idx, right_idx;
        patfile.read(&left_idx, sizeof(int));
        patfile.read(&right_idx, sizeof(int));
        node->left = node_table[left_idx];
        node->right = node_table[right_idx];

        assert(node->dump_idx == -1);
    }

//  streampos filesize = patfile.tellg();
    root_node = node_table[root_idx];

    return keyNum;
}


// Ρɤdumpƥե˳Ǽ
int
Patricia::save(ostream &patfile){
    patfile.seekp(0);

    if(root_node == NULL){
        cerr << "Patricia::save(): tree is empty." << endl;
        return -1;
    }
    Pat_node *node_table[keyNum];
    int total = dump_all(node_table, keyNum);

    // write header
    unsigned int magic = MAGIC;
    int root_idx = root_node->dump_idx;
    patfile.write(&magic, sizeof(magic));
    patfile.write(&keyNum, sizeof(keyNum));
    patfile.write(&root_idx, sizeof(root_idx));

    // write body
    for(int i = 0; i < total; i++){
        Pat_node *node = node_table[i];
        int len = strlen(node->key);

        patfile.put((unsigned char)len);
        patfile.write(node->key, len); // may len be 0
        patfile.write(& node->content, sizeof(node->content));
        patfile.write(& node->checkbit, sizeof(node->checkbit));
        assert(node->left->dump_idx != -1);
        patfile.write(& node->left->dump_idx, sizeof(int));
        assert(node->right->dump_idx != -1);
        patfile.write(& node->right->dump_idx, sizeof(int));
    }
    patfile << flush;

    streampos filesize = patfile.tellp();

    return (int)filesize;
}


// ΡɤֹƤ
int
Patricia::dump_all(Pat_node *table[], int table_size){
    if(NULL == root_node){
        return -1;
    }
    memset(table, 0, table_size * sizeof(int));
    dump_all_counter = 0;
    (void)dump_all_recursive(table, root_node);

    return dump_all_counter;
}

// dump_all
int
Patricia::dump_all_recursive(Pat_node *table[], Pat_node *here)
{
    if(here->left->checkbit > here->checkbit) 
        dump_all_recursive(table, here->left);
    else{
        if(here->left->dump_idx == -1){
            here->left->dump_idx = dump_all_counter;
            table[dump_all_counter] = here->left;
            dump_all_counter++;
        }
    }

    if(here->right->checkbit > here->checkbit)
        dump_all_recursive(table, here->right);
    else{
        if(here->right->dump_idx == -1){
            here->right->dump_idx = dump_all_counter;
            table[dump_all_counter] = here->right;
            dump_all_counter++;
        }
    }

    return dump_all_counter;
}


// key  prefix ˤʤäƤõ (see also prefix_search())
// ǲϤʤɤѤǤ
// Ĥä prefix ο֤
int
Patricia::search_prefix(const char *key, Pat_node *matches[]){
    Pat_node *forw, *here;
    int checkbit_table[MAXKEYLENGTH + 1];
    int keylen, match_count;
    if(NULL == root_node) return 0; // not found

    keylen = make_checkbit_table(checkbit_table, key);
    match_count = 0;
    forw = root_node;
    do{
        here = forw;
        // here->checkbit  prefix¸߾Ƥ뤫
        for(int i = 0; i < keylen && i < MAXKEYLENGTH; i++){
            if(checkbit_table[i] == here->checkbit){ // Ƥ
                Pat_node *h, *f;
                f = here;
                do{             // ʬڤκõ
                    h = f; f = h->left;
                }while(h->checkbit < f->checkbit);
                if(is_prefix(f->key, key)){ // ۤȤ prefix 
                    // cout << "prefix: " << f->key << endl;
                    matches[match_count++] = f;
                }
                break;
            }
        }
        forw = getbit(key, here->checkbit) ? here->right : here->left;
    }while(here->checkbit < forw->checkbit);
    if(is_prefix(forw->key, key)){ // Ǹ prefix ?
        // cout << "prefix: " << forw->key << endl;
        matches[match_count++] = forw;
    }
    matches[match_count] = NULL; // ֿ
    return match_count;
}


// search_prefix() ΤλޥӥåȤ
int
Patricia::make_checkbit_table(int *table, const char *key){
    int n = 0;
    const char *cp;
    int d;

    for(cp = key; *cp && (cp - key) < MAXKEYLENGTH; cp++){
        d = find_diffbit(cp, "");
        assert(d != -1);
        table[n++] = (int)(cp - key) * 8 + d;
    }
    table[n] = -1;              // ֿ?
    return n;                   // ä顤keyĹ֤
}


// -1: ǤƱ¸ߤƤ0: 
int
Patricia::insert(const char *key, long content){
    int diffbit;
    Pat_node *bottom;

#ifdef PATRICIA_DEBUG_VERBOSE
    cerr << "insert: " << key << ", " << content << endl;
#endif
    if(NULL == root_node){      // ڤäݤʤΤ root_node 
        root_node = new Pat_node(key, content);
        root_node->left = root_node;
        root_node->right = root_node;
        root_node->checkbit = -1; // ͤϤǤ?
        keyNum++;
        return INSERTOK;
    }
    
    bottom = down_node(key, root_node); // ꤽʥΡɤޤǹߤ
    if(0 == strcmp(key, bottom->key)){
//      cerr << "exist: " << key << ", " << bottom->content << endl;
        return EDUPLICATE;
    }
    diffbit = find_diffbit(key, bottom->key);

    Pat_node *here, *forw, *new_node;
    forw = root_node;
    do{                         // ٤Ρɤޤǹߤ
        here = forw;
        forw = getbit(key, here->checkbit) ? here->right : here->left;
        assert(forw->checkbit != diffbit);
    }while(   here->checkbit < forw->checkbit // üޤãƤʤ
           && forw->checkbit < diffbit );     // սɤȴƤʤ
    new_node = new Pat_node(key, content);
    new_node->checkbit = diffbit;
    if(getbit(key, diffbit)){
        new_node->left = forw;
        new_node->right = new_node;
    }else{
        new_node->left = new_node;
        new_node->right = forw;
    }
#if 0
    // ۤȤϢǽʬʤΤ
    if(getbit(key, here->checkbit))
        here->right = new_node;
    else here->left = new_node;
#else
    // ۤΤäȤΨΤˤ롣
    // Ƥ getbit(key, here->checkbit)  (forw == here->right)
    // root (here->right == here->left) ˤʤ뤳Ȥ롣
    if(here->right == here->left)
        (getbit(key, here->checkbit) ? here->right : here->left) = new_node;
    else
        ((forw == here->right) ? here->right : here->left) = new_node;
#endif
    keyNum++;
    return INSERTOK;
}

#ifdef PATRICIA_DEBUG
int main(){
    long x;
    Pat_node *prefixes[64];
    int i;

    Patricia *pat = new Patricia;
#if 1
    char line[1024];
    char *cp;
    long lineno = 0;
    while(! cin.eof()){
        lineno++;
        cin >> line;
        if(line[0] == '\0') break;
        if(Patricia::EDUPLICATE == pat->insert(line, lineno))
            cerr << "already stored? : " << line << endl;
    }
#endif
    
    cout << "----------------------------------------" << endl;
    x = pat->search("and");
    cout << "and -> " << x << endl;
    x = pat->search("harada");
    cout << "harada -> " << x << endl;

    cout << "----------------------------------------" << endl;
    x = pat->search_prefix("ޤǤ", prefixes);
    cout << "search_prefix, ޤǤ -> " << x << endl;
    for(i = 0; i < x; i++){
        cout << " -> " << prefixes[i]->key 
            <<  ", "   << prefixes[i]->content << endl;
    }

    cout << "----------------------------------------" << endl;
    x = pat->prefix_search("路", prefixes);
    cout << "prefix_search, 路 -> " << x << endl;
    for(i = 0; i < x; i++){
        cout << " -> " << prefixes[i]->key 
            <<  ", "   << prefixes[i]->content << endl;
    }

    cout << "----------------------------------------" << endl;
    int size = pat->save("dic/MORPH.pat");
    cout << "write: " << size << " bytes" << endl;

    delete pat;
    cout << endl;
    cout << "allocate: " << allocated_node << endl;
    cout << "released: " << released_node << endl;

    return 0;
}
#endif
