// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: valueSet.h,v 0.9 1997-12-27 07:53:38+09 harada Exp $
//
#ifndef VALUESET_H_INCLUDED
#define VALUESET_H_INCLUDED

template <class T>
class valueSet {                // 380yen
    friend ostream& operator<<(ostream &stream, valueSet &set);
private:
    int N;
    T *values;
public:
    valueSet() : N(0), values(NULL){}
    valueSet(const valueSet &that);
    valueSet(const T *list, int listLen);
    ~valueSet(){ if(values) delete []values; }

    operator const T *() const { return values; }
    const T & operator[](int nth){ return values[nth]; }
    int size(){ return N; }

    valueSet & operator=(const T &x);
    valueSet & operator=(const valueSet &s);
    valueSet & append(const T &x);
    valueSet & append(const T *list, int listLen);
    valueSet & operator+=(const T &x){ return append(x); }
};

template <class T> ostream&
operator<<(ostream &stream, valueSet<T> &set){
    stream << "( ";
    for(int i = 0; i < set.N; i++)
        stream << set.values[i] << ' ';
    stream << ") ";
    return stream;
}

template <class T>
valueSet<T>::valueSet(const valueSet &that)
    : N (that.N)
{
    values = new T[N];
    for(int i = 0; i < N; i++) values[i] = that.values[i];
}

template <class T> valueSet<T> &
valueSet<T>::operator=(const valueSet &s){
    if(s.N > N){
        if(values) delete []values;
        if(s.N > 0) values = new T[s.N]; else values = NULL;
    }
    N = s.N;
    for(int i = 0; i < N; i++) values[i] = s.values[i];
    return *this;
}

template <class T> valueSet<T> &
valueSet<T>::operator=(const T &x){
    if(values) delete []values;
    values = new T[1];
    values[0] = x;
    N = 1;
    return *this;
}

template <class T> valueSet<T> &
valueSet<T>::append(const T &x){
    if(N == 0){
        values = new T[1];
        values[0] = x;
    }else{
        T *newvalues = new T[N + 1];
        for(int i = 0; i < N; i++) newvalues[i] = values[i];
        newvalues[N] = x;
        delete []values;
        values = newvalues;
    }
    N++;
    return *this;
}

template <class T> valueSet<T> &
valueSet<T>::append(const T *list, int listLen){
    if(listLen == 0)
        return *this;           // do nothing
    if(N == 0){
        values = new T[listLen];
        for(int i = 0; i < listLen; i++)
            values[i] = list[i];
    }else{
        T *newvalues = new T[N + listLen];
        for(int j = 0; j < N; j++) newvalues[j] = values[j];
        for(int k = 0; k < listLen; k++) newvalues[N + k] = list[k];
        delete []values;
        values = newvalues;
    }
    N += listLen;
    return *this;
}

#if 0
#include <iostream.h>

int main(){
    valueSet<long> v;
    long list[] = {123, 456, 0, 9};

    v = 1;
    v = 2;
    v.append(list, sizeof(list)/sizeof(list[0]));
  
    for(int i = 0; i < v.size(); i++){
        cout << "i = " << i << " : " << v[i] << endl;
    }
  
    return 0;
}
#endif

#endif
